/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* SCIM Applet
 *
 * Copyright (C) 2006 Takuro Ashie <ashie@homa.ne.jp>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
 #include <config.h> /* Internationalization */
#endif
#define Uses_C_STDIO
#define Uses_C_STDLIB
#define Uses_SCIM_SOCKET
#define Uses_SCIM_TRANSACTION
#define Uses_SCIM_TRANS_COMMANDS
#define Uses_SCIM_CONFIG
#define Uses_SCIM_CONFIG_MODULE
#define Uses_SCIM_DEBUG
#define Uses_SCIM_HELPER
#define Uses_SCIM_HELPER_MODULE
#define Uses_SCIM_HELPER_MANAGER

#include <scim.h>
#include "scim_stl_map.h"
#include "scim_ui_utils.h"
#include "scim_panel_bridge_trans_commands.h"

#include <string.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <panel-applet.h>

using namespace scim;

#define SCIM_CONFIG_PANEL_GTK_FONT                        "/Panel/Gtk/Font"
#define SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_FACTORY_ICON   "/Panel/Gtk/ToolBar/ShowFactoryIcon"
#define SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_FACTORY_NAME   "/Panel/Gtk/ToolBar/ShowFactoryName"
#define SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_STICK_ICON     "/Panel/Gtk/ToolBar/ShowStickIcon"
#define SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_HELP_ICON      "/Panel/Gtk/ToolBar/ShowHelpIcon"
#define SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_MENU_ICON      "/Panel/Gtk/ToolBar/ShowMenuIcon"
#define SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_PROPERTY_LABEL "/Panel/Gtk/ToolBar/ShowPropertyLabel"

#define SCIM_MENU_HELPER_LIST_PLACE_HOLDER                "/popups/popup/HelperListPlaceHolder"

struct PropertyInfo {
    Property   property;
    GtkWidget *widget;

    PropertyInfo () : widget (0) { }
};

typedef std::vector <PropertyInfo> PropertyRepository;

typedef struct _SCIMToolbarApplet {
    PanelApplet        *applet;
    PanelAppletOrient   orient;
    GtkWidget          *main_box;
    GtkWidget          *property_box;
    GtkWidget          *trademark;
    GtkWidget          *stick_button;
    GtkWidget          *factory_button;
    GtkWidget          *menu_button;
    GtkWidget          *help_button;
    PropertyRepository  frontend_property_repository;
    bool                window_sticked;
    bool                toolbar_visible;
} SCIMToolbarApplet;

static void           sync_toolbar_state        (void);
static void           resize_child              (GtkWidget         *widget,
                                                 SCIMToolbarApplet *toolbar);
static GtkWidget     *create_toolbar_widget     (SCIMToolbarApplet *toolbar);
static ScimDirection  panel_orient_to_direction (PanelAppletOrient  orient);

static ConfigModule      *_config_module                = 0;
static ConfigPointer      _config;

static SocketClient       _bridge_socket;

static String             _current_factory_name;
static String             _current_factory_icon;
static PropertyList       _current_properties;

static GIOChannel        *_bridge_iochannel             = 0;
static guint              _bridge_iochannel_read_source = 0;

static bool               _toolbar_show_factory_icon    = true;
static bool               _toolbar_show_factory_name    = false;
static bool               _toolbar_show_stick_icon      = false;
static bool               _toolbar_show_help_icon       = false;
static bool               _toolbar_show_menu_icon       = false;
static bool               _toolbar_show_property_label  = false;

std::vector<SCIMToolbarApplet*>  _toolbar_list;

static PangoFontDescription     *_default_font_desc     = 0;

static void
request_factory_menu (void)
{
    Transaction send;

    // FIXME! should use SCIM_TRANS_CMD_REQUEST
    send.put_command (SCIM_TRANS_CMD_PANEL_REQUEST_FACTORY_MENU);
    send.write_to_socket (_bridge_socket);
}

static void
request_command_menu (void)
{
    Transaction send;

    // FIXME! should use SCIM_TRANS_CMD_REQUEST
    send.put_command (SCIM_TRANS_CMD_PANEL_BRIDGE_REQUEST_COMMAND_MENU);
    send.write_to_socket (_bridge_socket);
}

static void
request_help (void)
{
    Transaction send;

    // FIXME! should use SCIM_TRANS_CMD_REQUEST
    send.put_command (SCIM_TRANS_CMD_PANEL_REQUEST_HELP);
    send.write_to_socket (_bridge_socket);
}

static void
toggle_window_stick (void)
{
    Transaction send;

    // FIXME! should use SCIM_TRANS_CMD_REQUEST
    send.put_command (SCIM_TRANS_CMD_PANEL_BRIDGE_TOGGLE_WINDOW_STICK);
    send.write_to_socket (_bridge_socket);
}

static void
toggle_toolbar_visible (void)
{
    Transaction send;

    // FIXME! should use SCIM_TRANS_CMD_REQUEST
    send.put_command (SCIM_TRANS_CMD_PANEL_BRIDGE_TOGGLE_TOOLBAR_VISIBLE);
    send.write_to_socket (_bridge_socket);
}

static void
trigger_property (const String &property)
{
    Transaction send;

    // FIXME! should use SCIM_TRANS_CMD_REQUEST
    send.put_command (SCIM_TRANS_CMD_TRIGGER_PROPERTY);
    send.put_data (property);
    send.write_to_socket (_bridge_socket);
}

static void
ui_property_activate_cb (GtkWidget      *widget,
                         gpointer        user_data)
{
    GtkWidget *submenu = (GtkWidget *) g_object_get_data (G_OBJECT (widget), "property_submenu");

    if (submenu) {
#if 0
#if GDK_MULTIHEAD_SAFE
        if (_current_screen)
            gtk_menu_set_screen (GTK_MENU (submenu), _current_screen);
#endif
#endif
        guint32 activate_time = gtk_get_current_event_time ();
#if 0
        _property_menu_activated = true;
#endif
        gtk_menu_popup (GTK_MENU (submenu), 0, 0, 0, 0, 1, activate_time);
        return;
    }

    gchar * key = (gchar *) g_object_get_data (G_OBJECT (widget), "property_key");

    if (key) {
#if 0
        int client = GPOINTER_TO_INT (user_data);

        if (client < 0)
            trigger_property (key);
        else
            trigger_helper_property (client, key);
#else
        trigger_property (key);
#endif
    }
}

static GtkWidget *
create_properties_node (PropertyRepository           &repository,
                        PropertyList::const_iterator  begin,
                        PropertyList::const_iterator  end,
                        int                           client,
                        int                           level,
                        ScimDirection                 direction)
{
    PropertyList::const_iterator it;
    PropertyList::const_iterator next;

    GtkWidget * node;
    PropertyInfo info;
    bool leaf = true;

    if (begin >= end) return 0;

    // If the level is zero, then create the this node as button,
    // otherwise create as a menu item.
    if (!level) {
        GtkWidget * label
            = scim_ui_utils_create_label (begin->get_label (),
                                          begin->get_icon (),
                                          0,
                                          _default_font_desc,
                                          !_toolbar_show_property_label,
                                          false,
                                          direction);

        node = gtk_button_new ();
#if 1
        scim_ui_utils_force_no_focus_padding (node);
#endif
        gtk_container_add (GTK_CONTAINER (node), label);
        gtk_button_set_relief (GTK_BUTTON (node), GTK_RELIEF_NONE);
    } else {
        gint width, height;
        gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &width, &height);
        GtkWidget * icon
            = scim_ui_utils_create_icon (begin->get_icon (),
                                         NULL,
                                         width,
                                         height,
                                         false);
        node = gtk_image_menu_item_new_with_label (begin->get_label ().c_str ());
        if (icon)
            gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (node), icon);
    }

    if (begin->visible ())
        gtk_widget_show (node);
    else
        gtk_widget_hide (node);

    gtk_widget_set_sensitive (node, begin->active ());

#if 0
    if (_tooltips && begin->get_tip ().length ())
        gtk_tooltips_set_tip (_tooltips, node,
                              begin->get_tip ().c_str (), NULL);
#endif

    g_object_set_data_full (G_OBJECT (node), "property_key",
                            g_strdup (begin->get_key ().c_str ()),
                            g_free);

    info.property = *begin;
    info.widget = node;

    repository.push_back (info);

    it = begin + 1;

    if (it != end) {
        GtkWidget * submenu = gtk_menu_new ();
        GtkWidget * child;
        int menu_item_idx = 0;

        // Create all leafs of the first child.
        while (it != end) {
            // Find all leafs of the first child.
            for (next = it + 1; next != end; ++ next)
                if (!next->is_a_leaf_of (*it)) break;

            child = create_properties_node (repository, it,
                                            next, client, level + 1,
                                            direction);
            if (child) {
                gtk_menu_shell_append (GTK_MENU_SHELL (submenu), child);
                g_object_set_data (G_OBJECT (child),
                                   "menu_item_idx",
                                   GINT_TO_POINTER (menu_item_idx));
                ++ menu_item_idx;
            }

            it = next;
        }

        // The node is a button, so attach the submenu as its data.
        if (!level) {
            g_object_set_data_full (G_OBJECT (node),
                                    "property_submenu",
                                    submenu,
                                    (void (*)(void*)) gtk_widget_destroy);

#if 0
            g_signal_connect (G_OBJECT (submenu),
                              "deactivate",
                              G_CALLBACK (ui_property_menu_deactivate_cb),
                              NULL);
#endif
        } else // The node is a menu item, so attach the submenu directly.
            gtk_menu_item_set_submenu (GTK_MENU_ITEM (node), submenu);

        leaf = false;
    }

    if (leaf || level == 0) {
#if 1
        g_signal_connect (G_OBJECT (node),
                          ((level > 0) ? "activate" : "clicked"),
                          G_CALLBACK (ui_property_activate_cb),
                          GINT_TO_POINTER (client));
#endif
    }

    return node;
}

static void
create_properties (GtkWidget *container,
                   PropertyRepository &repository,
                   const PropertyList &properties,
                   int client,
                   int level,
                   ScimDirection direction)
{

    PropertyList::const_iterator it;
    PropertyList::const_iterator next;
    PropertyList::const_iterator begin = properties.begin ();
    PropertyList::const_iterator end = properties.end ();

    int menu_item_idx = 0;

    if (begin == end) return;

    it = begin;
    next = begin + 1;

    while (it != end) {
        if (next == end || !next->is_a_leaf_of (*it)) {
            GtkWidget * node = create_properties_node (repository,
                                                       it,
                                                       next,
                                                       client,
                                                       level,
                                                       direction);
            if (node) {
                // The container is a hbox.
                if (!level) {
                    if (direction == SCIM_DIR_BTT)
                        gtk_box_pack_end (GTK_BOX (container), node,
                                          TRUE, TRUE, 0);
                    else
                        gtk_box_pack_start (GTK_BOX (container), node,
                                            TRUE, TRUE, 0);
                }
                // The container is a menu.
                else {
                    gtk_menu_shell_append (GTK_MENU_SHELL (container), node);
                    g_object_set_data (G_OBJECT (node), "menu_item_idx",
                                       GINT_TO_POINTER (menu_item_idx));
                    ++ menu_item_idx;
                }
            }
            it = next;
        }
        ++ next;
    }
}

static void
update_property (PropertyRepository &repository,
                 const Property     &property,
                 bool                show_property_label,
                 GtkTooltips        *tooltips,
                 ScimDirection       direction)
{
    PropertyRepository::iterator it = repository.begin ();

    for (; it != repository.end (); ++ it) {
        if (it->property != property)
            continue;

        if (!it->widget) break;

        if (it->property.get_label () != property.get_label () ||
            it->property.get_icon () != property.get_icon ())
        {
            if (GTK_IS_BUTTON (it->widget)) {
                GtkWidget *label
                    = scim_ui_utils_create_label (property.get_label (),
                                                  property.get_icon (),
                                                  0,
                                                  _default_font_desc,
                                                  !show_property_label,
                                                  false,
                                                  direction);
                GtkWidget *old = gtk_bin_get_child (GTK_BIN (it->widget));
                gtk_container_remove (GTK_CONTAINER (it->widget), old);
                gtk_container_add (GTK_CONTAINER (it->widget), label);

            } else if (GTK_IS_MENU_ITEM (it->widget)) {
                gint width, height;
                gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &width, &height);

                GtkWidget *menu = gtk_widget_get_parent (it->widget);
                int menu_item_idx = GPOINTER_TO_INT (
                    g_object_get_data (G_OBJECT (it->widget), "menu_item_idx"));

                GtkWidget *icon = scim_ui_utils_create_icon (
                    property.get_icon (), NULL, width, height, false);
                GtkWidget * new_item = gtk_image_menu_item_new_with_label (
                    property.get_label ().c_str ());

                gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (new_item),
                                               icon);

                GtkWidget *submenu = gtk_menu_item_get_submenu (
                    GTK_MENU_ITEM (it->widget));

                gtk_menu_item_set_submenu (GTK_MENU_ITEM (new_item), submenu);

                g_object_set_data_full (G_OBJECT (new_item), "property_key",
                                        g_strdup (property.get_key ().c_str ()),
                                        g_free);
                g_object_set_data (G_OBJECT (new_item), "menu_item_idx",
                                   GINT_TO_POINTER (menu_item_idx));

                gtk_widget_destroy (it->widget);

                it->widget = new_item;
                gtk_menu_shell_insert (GTK_MENU_SHELL (menu),
                                       new_item,
                                       menu_item_idx);
            }
        }

        if (property.visible ())
            gtk_widget_show (it->widget);
        else
            gtk_widget_hide (it->widget);

        gtk_widget_set_sensitive (it->widget, property.active ());

        if (tooltips && property.get_tip ().length ())
            gtk_tooltips_set_tip (tooltips, it->widget,
                                  property.get_tip ().c_str (), NULL);

        it->property = property;
        break;
    }
}

static void
register_frontend_properties (PropertyRepository &repository,
                              const PropertyList &properties,
                              GtkWidget          *properties_area,
                              ScimDirection       direction)
{
    bool same = true;

    PropertyList::const_iterator pit = properties.begin ();

    if (properties.size () == 0) {
        same = false;
    } else if (properties.size () == repository.size ()) {
        // Check if the properties are same as old ones.
        PropertyRepository::iterator it = repository.begin ();

        for (; it != repository.end (); ++it, ++pit) {
            if (it->property != *pit) {
                same = false;
                break;
            }
        }
    } else {
        same = false;
    }

    if (same) {
        // Only update the properties.
        for (pit = properties.begin (); pit != properties.end (); ++pit)
            update_property (repository,
                             *pit,
                             _toolbar_show_property_label,
                             NULL,
                             direction);

    } else {
        // Construct all properties.
        gtk_container_foreach (GTK_CONTAINER (properties_area),
                               (GtkCallback) gtk_widget_destroy,
                               NULL);
        repository.clear ();

        if (properties.size ()) {
            create_properties (properties_area,
                               repository,
                               properties,
                               -1,
                               0,
                               direction);
        }
    }
}

static void
update_factory_button (GtkWidget *factory_button, String name, String icon,
                       ScimDirection direction)
{

    if (!GTK_IS_BUTTON (factory_button))
        return;

    GtkWidget * newlabel = 0;

    if (_toolbar_show_factory_icon) {
        newlabel = scim_ui_utils_create_label (name,
                                               icon,
                                               0,
                                               _default_font_desc,
                                               !_toolbar_show_factory_name,
                                               false,
                                               direction);
    } else {
        newlabel = gtk_label_new (name.c_str ());
        if (_default_font_desc)
            gtk_widget_modify_font (newlabel, _default_font_desc);
        gtk_widget_show (newlabel);
    }

    if (newlabel) {
        GtkWidget * old = gtk_bin_get_child (GTK_BIN (factory_button));
        if (old)
            gtk_container_remove (GTK_CONTAINER (factory_button), old);
        gtk_container_add (GTK_CONTAINER (factory_button), newlabel);
    }

#if 0
    if (!GTK_WIDGET_VISIBLE (factory_button) && !_toolbar_hidden)
        gtk_widget_show (factory_button);

    if (_tooltips)
        gtk_tooltips_set_tip (_tooltips, factory_button, name.c_str (), NULL);
#endif
}

static void
socket_update_factory_info (Transaction &recv)
{
    String uuid, name, lang, icon;

    if (!recv.get_data (uuid) || !recv.get_data (name) ||
        !recv.get_data (lang) || !recv.get_data (icon))
    {
        return;
    }

    lang = scim_get_normalized_language (lang);
    _current_factory_name = name;
    _current_factory_icon = icon;

    std::vector<SCIMToolbarApplet*>::iterator it;
    for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
        // Remove all properties
        // FIXME! this code should be removed when "turn on" "turn off"
        // commands are implemented.
        PropertyList properties;
        register_frontend_properties (
            (*it)->frontend_property_repository,
            properties,
            (*it)->property_box,
            panel_orient_to_direction ((*it)->orient));

        // update button
        update_factory_button ((*it)->factory_button, name,
                               icon,
                               panel_orient_to_direction ((*it)->orient));
    }
}

static void
socket_register_property (Transaction &recv)
{
    PropertyList properties;

    recv.get_data (properties);
    _current_properties = properties;

    std::vector<SCIMToolbarApplet*>::iterator it;
    for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
        register_frontend_properties (
            (*it)->frontend_property_repository,
            properties,
            (*it)->property_box,
            panel_orient_to_direction ((*it)->orient));
        gtk_container_foreach (GTK_CONTAINER ((*it)->main_box),
                               (GtkCallback) resize_child,
                               (*it)); 
   }
}

static void
socket_update_property (Transaction &recv)
{
    Property property;

    if (!recv.get_data (property))
        return;

    std::vector<SCIMToolbarApplet*>::iterator it;
    for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
        update_property ((*it)->frontend_property_repository,
                         property,
                         _toolbar_show_property_label,
                         NULL,
                         panel_orient_to_direction ((*it)->orient));
    }
}

static void
socket_notify_window_stick (Transaction &recv)
{
    uint32 sticked;

    if (!recv.get_data (sticked))
        return;

    std::vector<SCIMToolbarApplet*>::iterator it;
    for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
        BonoboUIComponent *component;
        PanelApplet *applet = PANEL_APPLET ((*it)->applet);
        component = panel_applet_get_popup_component (applet);
        bonobo_ui_component_set_prop (component,
                                      "/commands/StickWindows",
                                      "state", sticked ? "1" : "0", NULL);

        GtkWidget *image = gtk_bin_get_child (GTK_BIN ((*it)->stick_button));
        gtk_container_remove (GTK_CONTAINER ((*it)->stick_button), image);

        image = scim_ui_utils_create_stick_icon (sticked);
        gtk_container_add (GTK_CONTAINER ((*it)->stick_button), image);

        (*it)->window_sticked = sticked;
    }
}

static void
socket_notify_toolbar_visible (Transaction &recv)
{
    uint32 visible;

    if (!recv.get_data (visible))
        return;

    std::vector<SCIMToolbarApplet*>::iterator it;
    for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
        BonoboUIComponent *component;
        PanelApplet *applet = PANEL_APPLET ((*it)->applet);
        component = panel_applet_get_popup_component (applet);
        bonobo_ui_component_set_prop (component,
                                      "/commands/HideToolbar",
                                      "state", visible ? "0" : "1", NULL);
        (*it)->toolbar_visible = visible;
    }
}

static void
socket_reload_config (Transaction &recv)
{
    if (!_config.null ()) _config->reload ();
}

static gboolean
bridge_iochannel_handler (GIOChannel *source,
                          GIOCondition condition,
                          gpointer user_data)
{
    if (condition == G_IO_IN) {
        Transaction recv;
        int cmd = 0;

        if (!_bridge_socket.is_connected () ||
            !recv.read_from_socket (_bridge_socket, scim_get_default_socket_timeout ()))
        {
            return FALSE;
        }

        while (recv.get_command (cmd)) {
            if (cmd == SCIM_TRANS_CMD_PANEL_UPDATE_FACTORY_INFO)
                socket_update_factory_info (recv);
            else if (cmd == SCIM_TRANS_CMD_REGISTER_PROPERTIES)
                socket_register_property (recv);
            else if (cmd == SCIM_TRANS_CMD_UPDATE_PROPERTY)
                socket_update_property (recv);
            else if (cmd == SCIM_TRANS_CMD_PANEL_BRIDGE_NOTIFY_WINDOW_STICK)
                socket_notify_window_stick (recv);
            else if (cmd == SCIM_TRANS_CMD_PANEL_BRIDGE_NOTIFY_TOOLBAR_VISIBLE)
                socket_notify_toolbar_visible (recv);
            else if (cmd == SCIM_TRANS_CMD_RELOAD_CONFIG)
                socket_reload_config (recv);
        }

    } else if (condition == G_IO_ERR || condition == G_IO_HUP) {
        return FALSE;
    }
    return TRUE;
}

static void
load_config (void)
{
    if (_default_font_desc) {
        pango_font_description_free (_default_font_desc);
        _default_font_desc = 0;
    }

    if (!_config.null ()) {
        String str;

        str = _config->read (String (SCIM_CONFIG_PANEL_GTK_FONT),
                             String ("default"));
        if (str != String ("default"))
            _default_font_desc = pango_font_description_from_string (str.c_str ());

        _toolbar_show_factory_icon =
            _config->read (String (SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_FACTORY_ICON),
                           _toolbar_show_factory_icon);

        _toolbar_show_factory_name =
            _config->read (String (SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_FACTORY_NAME),
                           _toolbar_show_factory_name);

        _toolbar_show_stick_icon =
            _config->read (String (SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_STICK_ICON),
                           _toolbar_show_stick_icon);

        _toolbar_show_help_icon =
            _config->read (String (SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_HELP_ICON),
                           _toolbar_show_help_icon);

        _toolbar_show_menu_icon =
            _config->read (String (SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_MENU_ICON),
                           _toolbar_show_menu_icon);

        _toolbar_show_property_label =
            _config->read (String (SCIM_CONFIG_PANEL_GTK_TOOLBAR_SHOW_PROPERTY_LABEL),
                           _toolbar_show_property_label);
    }
}

static void
ui_reset (void)
{
    std::vector<SCIMToolbarApplet*>::iterator it;
    for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
        SCIMToolbarApplet *toolbar = (*it);
        if (_toolbar_show_stick_icon)
            gtk_widget_show (toolbar->stick_button);
        else
            gtk_widget_hide (toolbar->stick_button);

        if (_toolbar_show_menu_icon)
            gtk_widget_show (toolbar->menu_button);
        else
            gtk_widget_hide (toolbar->menu_button);

        if (_toolbar_show_help_icon)
            gtk_widget_show (toolbar->help_button);
        else
            gtk_widget_hide (toolbar->help_button);
    }

    sync_toolbar_state ();
}

static void
config_reload_cb (const ConfigPointer &config)
{
    _config = config;
    load_config ();
    ui_reset ();
}

static bool
initialize_config (const String &config_name)
{
    if (config_name != "dummy") {
        //load config module
        _config_module = new ConfigModule (config_name);

        if (!_config_module || !_config_module->valid ()) {
            std::cerr << "Can not load " << config_name << " Config module.\n";
            return false;
        }

        //create config instance
        _config = _config_module->create_config ();
    } else {
        _config = new DummyConfig ();
    }

    _config->signal_connect_reload (slot (config_reload_cb));

    return true;
}

static void
helper_verb_cb (BonoboUIComponent *component,
                gpointer           menu,
                const gchar       *verbname)
{
    HelperManager helper_manager;
    // FIXME! Specify display arg.
    helper_manager.run_helper (verbname, _config->get_name (), "");
}

static void
merge_helper_list (SCIMToolbarApplet *toolbar)
{
    HelperManager helper_manager;

    BonoboUIComponent *component;
    BonoboUINode *node;
    component = panel_applet_get_popup_component (toolbar->applet);

    for (unsigned int i = 0; i < helper_manager.number_of_helpers (); i++) {
        HelperInfo info;
        helper_manager.get_helper_info (i, info);

        if (!(info.option & SCIM_HELPER_STAND_ALONE))
            continue;

        node = bonobo_ui_node_new ("menuitem");
        bonobo_ui_node_set_attr (node, "name", info.uuid.c_str ());
        bonobo_ui_node_set_attr (node, "verb", info.uuid.c_str ());
        bonobo_ui_node_set_attr (node, "label", info.name.c_str ());
        bonobo_ui_node_set_attr (node, "pixtype", "filename");
        bonobo_ui_node_set_attr (node, "pixname", info.icon.c_str ());
        bonobo_ui_node_set_attr (node, "tip", info.description.c_str ());
        bonobo_ui_component_set_tree (component,
                                      SCIM_MENU_HELPER_LIST_PLACE_HOLDER,
                                      node, NULL);
        bonobo_ui_component_add_verb (component, info.uuid.c_str (),
                                      helper_verb_cb, NULL);
        bonobo_ui_node_free (node);
    }

    if (helper_manager.number_of_helpers ()) {
        node = bonobo_ui_node_new ("separator");
        bonobo_ui_component_set_tree (component,
                                      SCIM_MENU_HELPER_LIST_PLACE_HOLDER,
                                      node, NULL);
        bonobo_ui_node_free (node);
    }
}

static bool
open_connection (void)
{
    if (_bridge_socket.is_connected ())
        return true;

    bool ret = _bridge_socket.connect (
        SocketAddress ("local:/tmp/scim-panel-bridge-socket"));
    if (ret) {
        uint32 key;
        ret = scim_socket_open_connection (key, "Applet", "PanelBridge",
                                           _bridge_socket, 1000);
    }
    if (ret) {
        // connected
        Transaction recv;
        int cmd = 0;

        if (!_bridge_socket.is_connected () ||
            !recv.read_from_socket (_bridge_socket, scim_get_default_socket_timeout ()))
        {
            return FALSE;
        }

        if (!recv.get_command (cmd)) {
            // FIXME! close connection?
            return FALSE;
        }

        if (cmd != SCIM_TRANS_CMD_PANEL_BRIDGE_INITIALIZE) {
            // FIXME! close connection?
            return FALSE;
        }

        String config_name;
        if (!recv.get_data (config_name)) {
            // FIXME! close connection?
            return FALSE;
        }

        if (_config.null () && config_name.size ()) {
            // FIXME! check return value?
            initialize_config (config_name);
            load_config ();
            ui_reset ();
        }

        // observ the socket
        _bridge_iochannel = g_io_channel_unix_new (_bridge_socket.get_id());
        _bridge_iochannel_read_source
            = g_io_add_watch (_bridge_iochannel,
                              G_IO_IN,
                              bridge_iochannel_handler,
                              NULL);

#if 0
        std::vector<SCIMToolbarApplet*>::iterator it;
        for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
            merge_helper_list ((*it));
        }
#endif
    }

    return ret;
}

static void
verb_cb (BonoboUIComponent *component,
         gpointer           menu,
         const gchar       *verbname)
{
    if (!strcmp (verbname, "ReloadConfig")) {
        Transaction send;
        send.put_command (SCIM_TRANS_CMD_REPLY);
        send.put_command (SCIM_TRANS_CMD_RELOAD_CONFIG);
        send.write_to_socket (_bridge_socket);
        //if (!_config.null ()) _config->reload ();
    } else if (!strcmp (verbname, "Help")) {
        request_help ();
    } else {
        g_warning ("Unknown bonobo command '%s'", verbname);
    }
}

static void
ui_event_cb (BonoboUIComponent *comp,
             const gchar       *verbname,
             Bonobo_UIComponent_EventType type,
             const gchar       *state_string,
             SCIMToolbarApplet *toolbar)
{
    bool state = !strcmp (state_string, "1") ? TRUE : FALSE;

    if (!strcmp (verbname, "StickWindows")) {
        if (state != toolbar->window_sticked)
            toggle_window_stick ();
    } else if (!strcmp (verbname, "HideToolbar")) {
        if (state != !toolbar->toolbar_visible)
            toggle_toolbar_visible ();
    } else {
        g_warning ("Unknown bonobo command '%s'", verbname);
    }
}

static const BonoboUIVerb applet_menu_verbs [] = {
    BONOBO_UI_UNSAFE_VERB ("ReloadConfig", verb_cb),
    BONOBO_UI_UNSAFE_VERB ("StickWindows", NULL),
    BONOBO_UI_UNSAFE_VERB ("HideToolbar",  NULL),
    BONOBO_UI_UNSAFE_VERB ("Help",         verb_cb),
    BONOBO_UI_VERB_END
};

static void
resize_child (GtkWidget *widget, SCIMToolbarApplet *toolbar)
{
    gint width = -1, height = -1;
    PanelAppletOrient orient = panel_applet_get_orient (toolbar->applet);
    switch (orient)	{
    case PANEL_APPLET_ORIENT_LEFT:
    case PANEL_APPLET_ORIENT_RIGHT:
        width = panel_applet_get_size (toolbar->applet);
        break;
    case PANEL_APPLET_ORIENT_UP:
    case PANEL_APPLET_ORIENT_DOWN:
        height = panel_applet_get_size (toolbar->applet);
        break;
    default:
        break;
    }
    gtk_widget_set_size_request (widget, width, height);

    if (GTK_IS_CONTAINER (widget))
        gtk_container_foreach (GTK_CONTAINER (widget),
                               (GtkCallback) resize_child,
                               toolbar);
}

static void
sync_toolbar_state (void)
{
    Transaction send;

    // FIXME! should use SCIM_TRANS_CMD_REQUEST
    send.put_command (SCIM_TRANS_CMD_PANEL_BRIDGE_REQUEST_SYNC);
    send.write_to_socket (_bridge_socket);
}

static void
orient_changed_cb (PanelApplet       *applet,
                   PanelAppletOrient  orient,
                   SCIMToolbarApplet *toolbar)
{
    if (panel_orient_to_direction (orient) !=
        panel_orient_to_direction (toolbar->orient))
    {
        toolbar->orient = orient;
        gtk_widget_destroy (toolbar->main_box);
        toolbar->frontend_property_repository.clear ();
        create_toolbar_widget (toolbar);
        //sync_toolbar_state ();

        // restore factory button
        std::vector<SCIMToolbarApplet*>::iterator it;
        for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++)
            update_factory_button ((*it)->factory_button,
                                   _current_factory_name,
                                   _current_factory_icon,
                                   panel_orient_to_direction ((*it)->orient));

        // restore frontend properties
        for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
            register_frontend_properties (
                (*it)->frontend_property_repository,
                _current_properties,
                (*it)->property_box,
                panel_orient_to_direction ((*it)->orient));
        }
    }

    gtk_container_foreach (GTK_CONTAINER (toolbar->main_box),
                           (GtkCallback) resize_child,
                           toolbar);
}
static void
change_pixel_size_cb (PanelApplet       *applet,
                      gint               size,
                      SCIMToolbarApplet *toolbar)
{
    gtk_container_foreach (GTK_CONTAINER (toolbar->main_box),
                           (GtkCallback) resize_child,
                           toolbar);
}

#if 0
static void
size_allocate_cb (PanelApplet *applet,
                  GtkAllocation *allocation,
                  gpointer data)
{
}
#endif

static void
applet_destroy_cb (GtkWidget *widget, SCIMToolbarApplet *toolbar)
{
    std::vector<SCIMToolbarApplet*>::iterator it;
    for (it = _toolbar_list.begin (); it != _toolbar_list.end (); it++) {
        if (*it == toolbar) {
            _toolbar_list.erase (it);
            break;
        }
    }
    g_free (toolbar);
}

static GtkWidget *
add_button (SCIMToolbarApplet *toolbar,
            const char        *icon_file,
            const char        *label_text)
{
    ScimDirection direction = panel_orient_to_direction (toolbar->orient);

    // button
    GtkWidget *button = gtk_button_new ();
    gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
    scim_ui_utils_force_no_focus_padding (button);
    gtk_container_set_resize_mode (GTK_CONTAINER (button),
                                   GTK_RESIZE_IMMEDIATE);

    // label
    GtkWidget *label = scim_ui_utils_create_label (label_text ? label_text : "",
                                                   icon_file  ? icon_file  : "",
                                                   NULL,
                                                   _default_font_desc,
                                                   !label_text ? true : false,
                                                   false,
                                                   direction);
    gtk_container_add (GTK_CONTAINER (button), label);


    // add into parent
    if (direction == SCIM_DIR_BTT)
        gtk_box_pack_end (GTK_BOX (toolbar->main_box), button,
                          FALSE, FALSE, 0);
    else
        gtk_box_pack_start (GTK_BOX (toolbar->main_box), button,
                            FALSE, FALSE, 0);
    gtk_widget_show (button);

    return button;
}

static gboolean
timeout_open_connection (gpointer data)
{
    SCIMToolbarApplet *toolbar = (SCIMToolbarApplet*) data;
    static int try_count = 0;

    if (_bridge_socket.is_connected () || open_connection ()) {
        sync_toolbar_state ();
        merge_helper_list (toolbar);
        return FALSE;
    }

    try_count++;

    if (try_count < 20)
        return TRUE;
    else
        return FALSE;
}

static ScimDirection
panel_orient_to_direction (PanelAppletOrient orient)
{
    switch (orient)	{
    case PANEL_APPLET_ORIENT_LEFT:
        return SCIM_DIR_TTB;
        break;
    case PANEL_APPLET_ORIENT_RIGHT:
        return SCIM_DIR_BTT;
        break;
    case PANEL_APPLET_ORIENT_UP:
    case PANEL_APPLET_ORIENT_DOWN:
    default:
        return SCIM_DIR_HORIZONTAL;
        break;
    }
}

static GtkWidget *
create_toolbar_widget (SCIMToolbarApplet *toolbar)
{
    ScimDirection direction = panel_orient_to_direction (toolbar->orient);

    // create widgets
    if (direction == SCIM_DIR_HORIZONTAL)
        toolbar->main_box = gtk_hbox_new (FALSE, 0); 
    else
        toolbar->main_box = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (toolbar->applet), toolbar->main_box);
    gtk_widget_show (toolbar->main_box);

    // trademark icon
    toolbar->trademark = scim_ui_utils_create_trademark_icon ();
    if (direction == SCIM_DIR_BTT)
        gtk_box_pack_end (GTK_BOX (toolbar->main_box), toolbar->trademark,
                          FALSE, FALSE, 0);
    else
        gtk_box_pack_start (GTK_BOX (toolbar->main_box), toolbar->trademark,
                            FALSE, FALSE, 0);
    gtk_widget_show (toolbar->trademark);

    // stick button and factory button
    toolbar->stick_button   = add_button (toolbar, SCIM_PIN_UP_ICON_FILE, NULL);
    toolbar->factory_button = add_button (toolbar,
                                          _current_factory_icon.c_str (),
                                          _toolbar_show_factory_name ?
                                          _current_factory_name.c_str () :
                                          NULL);

    // property box
    if (direction == SCIM_DIR_HORIZONTAL)
        toolbar->property_box = gtk_hbox_new (FALSE, 0); 
    else
        toolbar->property_box = gtk_vbox_new (FALSE, 0); 
    if (direction == SCIM_DIR_BTT)
        gtk_box_pack_end (GTK_BOX (toolbar->main_box), toolbar->property_box,
                          FALSE, FALSE, 0);
    else
        gtk_box_pack_start (GTK_BOX (toolbar->main_box), toolbar->property_box,
                            FALSE, FALSE, 0);
    gtk_widget_show (toolbar->property_box);

    // other buttons
    toolbar->menu_button = add_button (toolbar, SCIM_MENU_ICON_FILE, NULL);
    toolbar->help_button = add_button (toolbar, SCIM_HELP_ICON_FILE, NULL);

    // connect signals
    g_signal_connect (G_OBJECT (toolbar->stick_button),
                      "clicked",
                      G_CALLBACK (toggle_window_stick),
                      NULL);
    g_signal_connect (G_OBJECT (toolbar->factory_button),
                      "clicked",
                      G_CALLBACK (request_factory_menu),
                      NULL);
    g_signal_connect (G_OBJECT (toolbar->menu_button),
                      "clicked",
                      G_CALLBACK (request_command_menu),
                      NULL);
    g_signal_connect (G_OBJECT (toolbar->help_button),
                      "clicked",
                      G_CALLBACK (request_help),
                      NULL);

    // hide non-implemented buttons
    gtk_widget_hide (toolbar->trademark);
    if (!_toolbar_show_stick_icon)
        gtk_widget_hide (toolbar->stick_button);
    if (!_toolbar_show_menu_icon)
        gtk_widget_hide (toolbar->menu_button);
    if (!_toolbar_show_help_icon)
        gtk_widget_hide (toolbar->help_button);

    return toolbar->main_box;
}

static gboolean
scim_applet_fill (PanelApplet *applet,
                  const gchar *iid,
                  gpointer     data)
{
    if (!iid || strcmp (iid, "OAFIID:GNOME_SCIMApplet"))
        return FALSE;

    if (_current_factory_name.empty ())
        _current_factory_name = _("English/Keyboard");
    if (_current_factory_icon.empty ())
    _current_factory_icon = SCIM_KEYBOARD_ICON_FILE;

    // create struct
    SCIMToolbarApplet *toolbar = g_new0 (SCIMToolbarApplet, 1);
    _toolbar_list.push_back (toolbar);
    toolbar->applet          = applet;
    toolbar->orient          = panel_applet_get_orient (toolbar->applet);
    toolbar->main_box        = NULL;
    toolbar->property_box    = NULL;
    toolbar->trademark       = NULL;
    toolbar->stick_button    = NULL;
    toolbar->factory_button  = NULL;
    toolbar->menu_button     = NULL;
    toolbar->help_button     = NULL;
    toolbar->window_sticked  = false;
    toolbar->toolbar_visible = false;

    // create widgets
    create_toolbar_widget (toolbar);

    // setup panel
    g_signal_connect (G_OBJECT (toolbar->applet), "change-orient",
                      G_CALLBACK (orient_changed_cb),  toolbar);
	g_signal_connect (G_OBJECT (toolbar->applet), "change_size",
                      G_CALLBACK (change_pixel_size_cb), toolbar);
    g_signal_connect (G_OBJECT (toolbar->applet), "destroy",
                      G_CALLBACK (applet_destroy_cb), toolbar);

    panel_applet_set_flags (PANEL_APPLET (applet),
                            PANEL_APPLET_HAS_HANDLE);
    panel_applet_setup_menu_from_file (applet,
                                       NULL,
                                       "GNOME_SCIMApplet.xml",
                                       NULL,
                                       applet_menu_verbs,
                                       NULL);
    BonoboUIComponent *component;
    component = panel_applet_get_popup_component (applet);
    g_signal_connect (component, "ui-event",
                      G_CALLBACK (ui_event_cb), toolbar);

    // show toolbar
    gtk_container_foreach (GTK_CONTAINER (toolbar->main_box),
                           (GtkCallback) resize_child,
                           toolbar);
    gtk_widget_show (GTK_WIDGET (applet));

    // open connection
    if (_bridge_socket.is_connected () || open_connection ()) {
        sync_toolbar_state ();
        merge_helper_list (toolbar);
    } else {
        // FIXME! check duplicates
        gtk_timeout_add (scim_get_default_socket_timeout () + 3000,
                         timeout_open_connection, toolbar);
    }

    return TRUE;
}

PANEL_APPLET_BONOBO_FACTORY ("OAFIID:GNOME_SCIMApplet_Factory",
                             PANEL_TYPE_APPLET,
                             "GNOME_SCIMApplet",
                             0,
                             scim_applet_fill,
                             NULL)
