/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

#include "scim_ui_utils.h"

#include "icons/up.xpm"
#include "icons/down.xpm"
#include "icons/left.xpm"
#include "icons/right.xpm"
//#include "icons/setup.xpm"
#include "icons/help.xpm"
#include "icons/trademark.xpm"
#include "icons/pin-up.xpm"
#include "icons/pin-down.xpm"
#include "icons/menu.xpm"

#define TOOLBAR_ICON_SIZE                     16 

GdkPixbuf *
scim_ui_utils_scale_pixbuf (GdkPixbuf *pixbuf,
			    int        width,
			    int        height)
{
    if (pixbuf) {
        if (gdk_pixbuf_get_width (pixbuf) != width ||
            gdk_pixbuf_get_height (pixbuf) != height)
        {
            GdkPixbuf *dest = gdk_pixbuf_scale_simple (pixbuf, width, height,
                                                       GDK_INTERP_BILINEAR);
            g_object_unref (pixbuf);
            pixbuf = dest;
        }
    }
    return pixbuf;
}

GtkWidget *
scim_ui_utils_create_icon (const String  &iconfile,
                           const char   **xpm,
                           int            width,
                           int            height,
                           bool           force_create)
{
    String path = iconfile;
    GdkPixbuf *pixbuf = 0;

    if (path.length ()) {
        // Not a absolute path, prepend SCIM_ICONDIR
        if (path [0] != SCIM_PATH_DELIM)
            path = String (SCIM_ICONDIR) + String (SCIM_PATH_DELIM_STRING) + path;

        pixbuf = gdk_pixbuf_new_from_file (path.c_str (), 0);
    }

    if (!pixbuf && xpm) {
        pixbuf = gdk_pixbuf_new_from_xpm_data (xpm);
    }

    if (!pixbuf && force_create) {
        if (width <= 0 || height <= 0)
            return 0;

        pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, true, 8, width, height);

        if (!pixbuf)
            return 0;

        gdk_pixbuf_fill (pixbuf, 0);
    }

    if (pixbuf) {
        if (width <= 0) width = gdk_pixbuf_get_width (pixbuf);
        if (height <= 0) height = gdk_pixbuf_get_height (pixbuf);

        pixbuf = scim_ui_utils_scale_pixbuf (pixbuf, width, height);

        GtkWidget *icon = gtk_image_new_from_pixbuf (pixbuf);
        gtk_widget_show (icon);

        gdk_pixbuf_unref (pixbuf);

        return icon;
    }
    return 0;
}

GtkWidget *
scim_ui_utils_create_label (const String          &name,
                            const String          &iconfile,
                            const char           **xpm,
                            PangoFontDescription  *font_desc,
                            bool                   show_icon_only,
                            bool                   force_icon,
                            ScimDirection          direction)
{
    GtkWidget * box;
    if (direction == SCIM_DIR_HORIZONTAL)
        box = gtk_hbox_new (FALSE, 0);
    else
        box = gtk_vbox_new (FALSE, 0);

    GtkWidget * label = gtk_label_new (name.c_str ());
    if (direction == SCIM_DIR_BTT)
        gtk_label_set_angle (GTK_LABEL (label), 90.0);
    else if (direction == SCIM_DIR_TTB)
        gtk_label_set_angle (GTK_LABEL (label), 270.0);

    gint width, height;

    if (font_desc)
        gtk_widget_modify_font (label, font_desc);

    gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &width, &height);

    GtkWidget *icon = scim_ui_utils_create_icon (iconfile,
                                                 xpm,
                                                 width,
                                                 height,
                                                 force_icon);

    if (icon) {
        if (direction == SCIM_DIR_BTT)
            gtk_box_pack_end (GTK_BOX (box), icon, FALSE, FALSE, 0);
        else
            gtk_box_pack_start (GTK_BOX (box), icon, FALSE, FALSE, 0);
        if (!show_icon_only)
            gtk_box_set_spacing (GTK_BOX (box), 4);
    }

    if (!show_icon_only || !icon) {
        if (direction == SCIM_DIR_BTT)
            gtk_box_pack_end (GTK_BOX (box), label, FALSE, FALSE, 0);
        else
            gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
    } else {
        gtk_widget_destroy (label);
    }

    gtk_widget_show_all (box);

    return box;
}

void
scim_ui_utils_force_no_focus_padding (GtkWidget *widget)
{
    gtk_rc_parse_string (
        "\n"
        "   style \"scim-applet-button-style\"\n"
        "   {\n"
        "      GtkWidget::focus-line-width=0\n"
        "      GtkWidget::focus-padding=0\n"
        "   }\n"
        "\n"
        "    widget \"*.scim-applet-button\" style \"scim-applet-button-style\"\n"
        "\n");
    gtk_widget_set_name (widget, "scim-applet-button");
}

GtkWidget *
scim_ui_utils_create_trademark_icon (void)
{
    return scim_ui_utils_create_icon (SCIM_TRADEMARK_ICON_FILE,
                                      (const char **) trademark_xpm,
                                      TOOLBAR_ICON_SIZE + 4,
                                      TOOLBAR_ICON_SIZE + 4);
}

GtkWidget *
scim_ui_utils_create_stick_icon (bool sticked)
{
    return scim_ui_utils_create_icon (
        (sticked ? SCIM_PIN_DOWN_ICON_FILE : SCIM_PIN_UP_ICON_FILE),
        (const char **) (sticked ? pin_down_xpm : pin_up_xpm),
        TOOLBAR_ICON_SIZE,
        TOOLBAR_ICON_SIZE);
}

GtkWidget *
scim_ui_utils_create_help_icon (void)
{
    return scim_ui_utils_create_icon (SCIM_HELP_ICON_FILE,
                                      (const char **) help_xpm,
                                      TOOLBAR_ICON_SIZE,
                                      TOOLBAR_ICON_SIZE);
}

GtkWidget *
scim_ui_utils_create_menu_icon (void)
{
    return scim_ui_utils_create_icon (SCIM_MENU_ICON_FILE,
                                      (const char **) menu_xpm,
                                      TOOLBAR_ICON_SIZE,
                                      TOOLBAR_ICON_SIZE);
}

GtkWidget *
scim_ui_utils_create_up_icon (void)
{
    return scim_ui_utils_create_icon (SCIM_UP_ICON_FILE,
                                      (const char **) up_xpm,
                                      TOOLBAR_ICON_SIZE,
                                      TOOLBAR_ICON_SIZE);
}

GtkWidget *
scim_ui_utils_create_left_icon (void)
{
    return scim_ui_utils_create_icon (SCIM_LEFT_ICON_FILE,
                                      (const char **) left_xpm,
                                      TOOLBAR_ICON_SIZE,
                                      TOOLBAR_ICON_SIZE);
}

GtkWidget *
scim_ui_utils_create_right_icon (void)
{
    return scim_ui_utils_create_icon (SCIM_RIGHT_ICON_FILE,
                                      (const char **) right_xpm,
                                      TOOLBAR_ICON_SIZE,
                                      TOOLBAR_ICON_SIZE);
}

GtkWidget *
scim_ui_utils_create_down_icon (void)
{
    return scim_ui_utils_create_icon (SCIM_DOWN_ICON_FILE,
                                      (const char **) down_xpm,
                                      TOOLBAR_ICON_SIZE,
                                      TOOLBAR_ICON_SIZE);
}
